package org.nakedobjects.nof.reflect.remote;

import org.apache.log4j.Logger;
import org.nakedobjects.noa.NakedObjectRuntimeException;
import org.nakedobjects.noa.adapter.Naked;
import org.nakedobjects.noa.adapter.NakedObject;
import org.nakedobjects.noa.adapter.NakedReference;
import org.nakedobjects.noa.adapter.Oid;
import org.nakedobjects.noa.persist.ConcurrencyException;
import org.nakedobjects.noa.reflect.NakedObjectAction;
import org.nakedobjects.noa.spec.Features;
import org.nakedobjects.noa.spec.NakedObjectSpecification;
import org.nakedobjects.nof.core.context.NakedObjectsContext;
import org.nakedobjects.nof.reflect.peer.AbstractActionPeer;
import org.nakedobjects.nof.reflect.peer.ActionPeer;
import org.nakedobjects.nof.reflect.peer.MemberIdentifier;
import org.nakedobjects.nof.reflect.peer.ReflectiveActionException;
import org.nakedobjects.nof.reflect.remote.data.Data;
import org.nakedobjects.nof.reflect.remote.data.Distribution;
import org.nakedobjects.nof.reflect.remote.data.KnownObjects;
import org.nakedobjects.nof.reflect.remote.data.NullData;
import org.nakedobjects.nof.reflect.remote.data.ObjectData;
import org.nakedobjects.nof.reflect.remote.data.ObjectEncoder;
import org.nakedobjects.nof.reflect.remote.data.ReferenceData;
import org.nakedobjects.nof.reflect.remote.data.ServerActionResultData;


/**
 * A reflection peer for executing actions remotely, instead of on the local machine. Any calls to
 * <code>execute</code> are passed over the network to the server for invocation. There are two cases where
 * the request is not passed to the server, ie it is executed locally: 1) where the method is static, ie is on
 * the class rather than an instance; 2) if the instance is not persistent; 3) if the method is marked as
 * 'local'. If a method is marked as being 'remote' then static methods and methods on transient objects will
 * be passed to the server.
 * 
 * <p>
 * If any of the objects involved have been changed on the server by another process then a
 * ConcurrencyException will be passed back to the client and re-thrown.
 * </p>
 */
final class ProxyAction extends AbstractActionPeer {
    private final static Logger LOG = Logger.getLogger(ProxyAction.class);
    private Distribution connection;
    private final ObjectEncoder encoder;

    public ProxyAction(final ActionPeer local, final Distribution connection, final ObjectEncoder encoder) {
        super(local);
        this.connection = connection;
        this.encoder = encoder;
    }

    private String debug(
            final String message,
            final MemberIdentifier identifier,
            final NakedReference target,
            final Naked[] parameters) {
        if (LOG.isDebugEnabled()) {
            StringBuffer str = new StringBuffer();
            str.append(message);
            str.append(" ");
            str.append(identifier);
            str.append(" on ");
            str.append(target);
            for (int i = 0; i < parameters.length; i++) {
                if (i > 0) {
                    str.append(',');
                }
                str.append(parameters[i]);
            }
            return str.toString();
        } else {
            return "";
        }
    }

    public Naked execute(final NakedReference target, final Naked[] parameters) throws ReflectiveActionException {
        if (isToBeExecutedRemotely(target)) {
            /*
             * NOTE - only remotely executing actions on objects not collection - due to collections not
             * having OIDs yet
             */
            return executeRemotely((NakedObject) target, parameters);
        } else {
            LOG.debug(debug("execute locally", getIdentifier(), target, parameters));
            return super.execute(target, parameters);
        }
    }

    private Naked executeRemotely(final NakedObject target, final Naked[] parameters) {
        KnownObjects knownObjects = new KnownObjects();
        Data[] parameterObjectData = parameterValues(parameters, knownObjects);
        LOG.debug(debug("execute remotely", getIdentifier(), target, parameters));
        ReferenceData targetReference = target == null ? null : encoder.createActionTarget(target, knownObjects);
        ServerActionResultData result;
        try {
            String name = getIdentifier().getClassName() + "#" + getIdentifier().getName();
            result = connection.executeServerAction(NakedObjectsContext.getSession(), getType().getName(), name, targetReference,
                    parameterObjectData);
        } catch (ConcurrencyException e) {
            Oid source = e.getSource();
            if (source == null) {
                throw e;
            } else {
                NakedObject failedObject = NakedObjectsContext.getObjectLoader().getAdapterFor(source);
                NakedObjectsContext.getObjectPersistor().reload(failedObject);
                LOG.info("concurrency conflict: " + e.getMessage());
                throw new ConcurrencyException("Object automatically reloaded: " + failedObject.titleString(), e);
            }
        } catch (NakedObjectRuntimeException e) {
            LOG.error("remoting exception", e);
            throw e;
        }

        // must deal with transient-now-persistent objects first
        if (target.getResolveState().isTransient()) {
            encoder.madePersistent(target, result.getPersistedTarget());
        }

        for (int i = 0; i < parameters.length; i++) {
            if (getParameterTypes()[i].getType() == NakedObjectSpecification.OBJECT) {
                encoder.madePersistent((NakedObject) parameters[i], result.getPersistedParameters()[i]);
            }
        }

        Data returned = result.getReturn();
        Naked returnedObject;
        returnedObject = returned instanceof NullData ? null : encoder.restore(returned);

        ObjectData[] updates = result.getUpdates();
        for (int i = 0; i < updates.length; i++) {
            LOG.debug("update " + updates[i].getOid());
            encoder.restore(updates[i]);
        }


        ReferenceData[] disposed = result.getDisposed();
        for (int i = 0; i < disposed.length; i++) {
            Oid oid = disposed[i].getOid();
            LOG.debug("disposed " + oid);
            NakedObject adapter = NakedObjectsContext.getObjectLoader().getAdapterFor(oid);
            NakedObjectsContext.getUpdateNotifer().addDisposedObject(adapter);
        }

        String[] messages = result.getMessages();
        for (int i = 0; i < messages.length; i++) {
            NakedObjectsContext.getMessageBroker().addMessage(messages[i]);
        }

        messages = result.getWarnings();
        for (int i = 0; i < messages.length; i++) {
            NakedObjectsContext.getMessageBroker().addWarning(messages[i]);
        }

        return returnedObject;
    }

    private boolean isToBeExecutedRemotely(final NakedReference target) {
        boolean remoteOverride = getTarget() == NakedObjectAction.REMOTE;
        boolean localOverride = getTarget() == NakedObjectAction.LOCAL;

        if (localOverride) {
            return false;
        }

        if (remoteOverride) {
            return true;
        }

        if (Features.isService(target.getSpecification())) {
            return true;
        }
        
        if (target == null) {
            // for static methods there is no target
            return false;
        }

        boolean remoteAsPersistent = target.getResolveState().isPersistent();
        return remoteAsPersistent;
    }

    private Data[] parameterValues(final Naked[] parameters, final KnownObjects knownObjects) {
        NakedObjectSpecification[] parameterTypes = getParameterTypes();
        return encoder.createParameters(parameterTypes, parameters, knownObjects);
    }
}
// Copyright (c) Naked Objects Group Ltd.
