package org.nakedobjects.nof.reflect.remote;

import org.apache.log4j.Logger;
import org.nakedobjects.noa.adapter.NakedObject;
import org.nakedobjects.noa.adapter.Oid;
import org.nakedobjects.noa.persist.ConcurrencyException;
import org.nakedobjects.nof.core.context.NakedObjectsContext;
import org.nakedobjects.nof.reflect.peer.AbstractOneToOnePeer;
import org.nakedobjects.nof.reflect.peer.OneToOnePeer;
import org.nakedobjects.nof.reflect.remote.data.Distribution;
import org.nakedobjects.nof.reflect.remote.data.IdentityData;
import org.nakedobjects.nof.reflect.remote.data.ObjectData;
import org.nakedobjects.nof.reflect.remote.data.ObjectEncoder;


/**
 * A reflection peer for changing one-to-one fields remotely, instead of on the local machine. Any requests
 * to set or clear an association will be passed over the network to the server for completion. Only
 * requests on persistent objects are passed to the server; on a transient object the request will always be
 * dealt with locally.
 * 
 * <p>
 * If either of the objects involved have been changed on the server by another process then a
 * ConcurrencyException will be passed back to the client and re-thrown.
 * </p>
 */
final class ProxyOneToOneAssociation extends AbstractOneToOnePeer {
    private final static Logger LOG = Logger.getLogger(ProxyOneToOneAssociation.class);
    private final Distribution connection;
    private final ObjectEncoder encoder;

    public ProxyOneToOneAssociation(final OneToOnePeer local, final Distribution connection, final ObjectEncoder objectDataFactory) {
        super(local);
        this.connection = connection;
        this.encoder = objectDataFactory;
    }

    public void clearAssociation(final NakedObject inObject, final NakedObject associate) {
        if (executeRemotely(inObject)) {
            LOG.debug("clear association remotely " + inObject + "/" + associate);
            IdentityData targetReference = encoder.createIdentityData(inObject);
            IdentityData associateReference = encoder.createIdentityData(associate);
            ObjectData[] updates;
            try {
                updates = connection.clearAssociation(NakedObjectsContext.getSession(), getIdentifier().getName(),
                        targetReference, associateReference);
            } catch (ConcurrencyException e) {
                throw concurrencyException(e);
            }
            updateChangedObjects(updates);
        } else {
            LOG.debug("clear association locally " + inObject + "/" + associate);
            super.clearAssociation(inObject, associate);
        }
    }

    private ConcurrencyException concurrencyException(final ConcurrencyException e) {
        LOG.info("concurrency conflict: " + e.getMessage());
        Oid source = e.getSource();
        if (source == null) {
            return e;
        } else {
            NakedObject failedObject = NakedObjectsContext.getObjectLoader().getAdapterFor(source);
            NakedObjectsContext.getObjectPersistor().reload(failedObject);
            return new ConcurrencyException("Object automatically reloaded: " + failedObject.titleString(), e);
        }
    }

    private boolean executeRemotely(final NakedObject target) {
        return target.getResolveState().isPersistent();
    }

    public void setAssociation(final NakedObject inObject, final NakedObject associate) {
        if (executeRemotely(inObject)) {
            LOG.debug("set association remotely " + getIdentifier() + " in " + inObject + " with " + associate);
            IdentityData targetReference = encoder.createIdentityData(inObject);
            IdentityData associateReference = encoder.createIdentityData(associate);
            ObjectData[] updates;
            try {
                updates = connection.setAssociation(NakedObjectsContext.getSession(), getIdentifier().getName(), targetReference,
                        associateReference);
            } catch (ConcurrencyException e) {
                throw concurrencyException(e);
            }
            updateChangedObjects(updates);
        } else {
            LOG.debug("set association locally " + getIdentifier() + " in " + inObject + " with " + associate);
            super.setAssociation(inObject, associate);
        }
    }

    private void updateChangedObjects(final ObjectData[] updates) {
        for (int i = 0; i < updates.length; i++) {
            LOG.debug("update " + updates[i].getOid());
            encoder.restore(updates[i]);
        }
    }
}
// Copyright (c) Naked Objects Group Ltd.
