package org.nakedobjects.nof.reflect.remote.data;

import org.nakedobjects.noa.adapter.Oid;
import org.nakedobjects.noa.adapter.Version;
import org.nakedobjects.noa.reflect.NakedObjectField;
import org.nakedobjects.noa.security.Session;
import org.nakedobjects.noa.spec.NakedObjectSpecification;
import org.nakedobjects.nof.core.context.NakedObjectsContext;
import org.nakedobjects.nof.core.util.Logger;

import java.util.Properties;
import java.util.Vector;


public class DistributionLogger extends Logger implements Distribution {
    private String padding = "      ";
    private final ObjectEncoder encoder;

    private String dump(final Data data) {
        StringBuffer str = new StringBuffer();
        dump(str, data, 1, new Vector());
        return str.toString();
    }

    private String dump(final Data[] data) {
        StringBuffer str = new StringBuffer();
        for (int i = 0; i < data.length; i++) {
            str.append("\n    [");
            str.append(i + 1);
            str.append("] ");
            dump(str, data[i], 3, new Vector());
        }
        return str.toString();
    }

    private void dump(final StringBuffer str, final Data data, final int indent, final Vector complete) {
        if (data == null) {
            str.append("null");
        } else if (data instanceof NullData) {
            str.append("NULL (NullData object)");
        } else if (data instanceof ValueData) {
            ValueData valueData = ((ValueData) data);
            str.append("ValueData@" + Integer.toHexString(valueData.hashCode()) + " " + valueData.getType() + ":"
                    + valueData.getEncodedValue());
        } else if (data instanceof IdentityData) {
            IdentityData referenceData = (IdentityData) data;
            str.append("ReferenceData@" + Integer.toHexString(referenceData.hashCode()) + " " + referenceData.getType() + ":"
                    + referenceData.getOid() + ":" + referenceData.getVersion());
        } else if (data instanceof ObjectData) {
            dumpObjectData(str, data, indent, complete);
        } else if (data instanceof CollectionData) {
            dumpCollectionData(str, data, indent, complete);
        } else {
            str.append("Unknown: " + data);
        }
    }

    private void dumpCollectionData(final StringBuffer str, final Data data, final int indent, final Vector complete) {
        CollectionData objectData = ((CollectionData) data);
        str.append("CollectionData@" + Integer.toHexString(objectData.hashCode()) + " " + objectData.getType() + ":"
                + objectData.getOid() + ":" + (objectData.hasAllElements() ? "A" : "-") + ":" + objectData.getVersion());
        Object[] elements = objectData.getElements();
        for (int i = 0; elements != null && i < elements.length; i++) {
            str.append("\n");
            str.append(padding(indent));
            str.append(i + 1);
            str.append(") ");
            dump(str, (Data) elements[i], indent + 1, complete);
        }
    }

    private void dumpObjectData(final StringBuffer str, final Data data, final int indent, final Vector complete) {
        ObjectData objectData = ((ObjectData) data);
        str.append("ObjectData@" + Integer.toHexString(objectData.hashCode()) + " " + objectData.getType() + ":"
                + objectData.getOid() + ":" + (objectData.hasCompleteData() ? "C" : "-") + ":" + objectData.getVersion());

        if (complete.contains(objectData)) {
            str.append(" (already detailed)");
            return;
        }

        complete.addElement(objectData);
        NakedObjectSpecification spec = NakedObjectsContext.getReflector().loadSpecification(data.getType());
        NakedObjectField[] fs = encoder.getFieldOrder(spec);
        Object[] fields = objectData.getFieldContent();
        for (int i = 0; fields != null && i < fields.length; i++) {
            str.append("\n");
            str.append(padding(indent));
            str.append(i + 1);
            str.append(") ");
            str.append(fs[i].getId());
            str.append(": ");
            dump(str, (Data) fields[i], indent + 1, complete);
        }
    }

    private String indentedNewLine() {
        return "\n" + padding(2);
    }

    public void init() {
        decorated.init();
    }

    private String padding(final int indent) {
        int length = indent * 3;
        while (length > padding.length()) {
            padding += padding;
        }
        return padding.substring(0, length);
    }

    private final Distribution decorated;

    public DistributionLogger(final ObjectEncoder encoder, final Distribution decorated, final String fileName) {
        super(fileName, false);
        this.encoder = encoder;
        this.decorated = decorated;
    }

    public DistributionLogger(final ObjectEncoder encoder, final Distribution decorated) {
        super(null, true);
        this.encoder = encoder;
        this.decorated = decorated;
    }

    public Session authenticate(String data) {
        log("authenticate");
        return decorated.authenticate(data);
    }

    public boolean authoriseUsability(final Session session, final String data) {
        log("authoriseUsability");
        return decorated.authoriseUsability(session, data);
    }

    public boolean authoriseVisibility(final Session session, final String data) {
        log("authoriseVisibility");
        return decorated.authoriseVisibility(session, data);
    }

    public ObjectData[] clearAssociation(
            final Session session,
            final String fieldIdentifier,
            final IdentityData target,
            final IdentityData associate) {
        log("clear association " + fieldIdentifier + indentedNewLine() + "target: " + dump(target) + indentedNewLine()
                + "associate: " + dump(associate));
        ObjectData[] changes = decorated.clearAssociation(session, fieldIdentifier, target, associate);
        log("  <-- changes: " + dump(changes));
        return changes;
    }

    public ObjectData[] clearValue(final Session session, final String fieldIdentifier, final IdentityData target) {
        log("clear value " + fieldIdentifier + indentedNewLine() + "target: " + dump(target));
        ObjectData[] changes = decorated.clearValue(session, fieldIdentifier, target);
        log("  <-- changes: " + dump(changes));
        return changes;
    }

    public void closeSession(Session session) {
        log("close session " + session);
        decorated.closeSession(session);
    }

    public ServerActionResultData executeServerAction(
            final Session session,
            final String actionType,
            final String actionIdentifier,
            final ReferenceData target,
            final Data[] parameters) {
        log("execute action " + actionIdentifier + "/" + actionType + indentedNewLine() + "target: " + dump(target)
                + indentedNewLine() + "parameters: " + dump(parameters));
        ServerActionResultData result;
        try {
            result = decorated.executeServerAction(session, actionType, actionIdentifier, target, parameters);
            log("  <-- returns: " + dump(result.getReturn()));
            log("  <-- persisted target: " + dump(result.getPersistedTarget()));
            log("  <-- persisted parameters: " + dump(result.getPersistedParameters()));
            log("  <-- updates: " + dump(result.getUpdates()));
            log("  <-- disposed: " + dump(result.getDisposed()));
        } catch (RuntimeException e) {
            log("  <-- exception: " + e.getClass().getName() + " " + e.getMessage());
            throw e;
        }
        return result;
    }

    public ClientActionResultData executeClientAction(final Session session, final ReferenceData[] data, final int[] types) {
        Vector complete = new Vector();
        StringBuffer str = new StringBuffer();
        for (int i = 0; i < data.length; i++) {
            str.append(indentedNewLine());
            str.append("[");
            str.append(i + 1);
            str.append("] ");
            switch (types[i]) {
            case ClientTransactionEvent.ADD:
                str.append("persisted: ");
                break;
            case ClientTransactionEvent.CHANGE:
                str.append("changed: ");
                break;
            case ClientTransactionEvent.DELETE:
                str.append("deleted: ");
                break;
            }
            dump(str, data[i], 3, complete);
        }
        log("execute client action " + str);

        /*
         * + indentedNewLine() + "changed: " + dump(changed, complete) + indentedNewLine() + "deleted: " +
         * dump(deleted, complete));
         */
        ClientActionResultData results = decorated.executeClientAction(session, data, types);

        complete = new Vector();
        str = new StringBuffer();
        ReferenceData[] persistedUpdates = results.getPersisted();
        Version[] changedVersions = results.getChanged();
        for (int i = 0; i < persistedUpdates.length; i++) {
            str.append(indentedNewLine());
            str.append("[");
            str.append(i + 1);
            str.append("] ");
            switch (types[i]) {
            case ClientTransactionEvent.ADD:
                str.append("persisted: ");
                dump(str, persistedUpdates[i], 3, complete);
                break;
            case ClientTransactionEvent.CHANGE:
                str.append("changed: ");
                str.append(changedVersions[i]);
                break;
            }
        }
        log(" <--- execute client action results" + str);
        /*
         * log(" <-- persisted: " + dump(results.getPersisted())); log(" <-- changed: " +
         * dump(results.getChanged()));
         */
        return results;
    }

    public ObjectData[] findInstances(final Session session, final CriteriaData criteria) {
        log("find instances " + criteria);
        ObjectData[] instances = decorated.findInstances(session, criteria);
        log(" <-- instances: " + dump(instances));
        return instances;
    }

    protected Class getDecoratedClass() {
        return decorated.getClass();
    }

    public ObjectData getObject(Session session, final Oid oid, final String specificationName) {
        log("get object " + oid);
        ObjectData data = decorated.getObject(null, oid, specificationName);
        log(" <-- data: " + data);
        return data;
    }

    public Properties getProperties() {
        log("get properties");
        Properties properties = decorated.getProperties();
        log(" <-- data: " + properties);
        return properties;
    }

    public boolean hasInstances(final Session session, final String fullName) {
        log("has instances " + fullName);
        boolean hasInstances = decorated.hasInstances(session, fullName);
        log(" <-- instances: " + (hasInstances ? "yes" : "no"));
        return hasInstances;
    }

    public IdentityData oidForService(Session session, String id) {
        log("oid for resource " + id);
        IdentityData result = decorated.oidForService(session, id);
        log(" <-- data: " + dump(result));
        return result;
    }

    public Data resolveField(final Session session, final IdentityData data, final String name) {
        log("resolve field " + name + " - " + dump(data));
        Data result = decorated.resolveField(session, data, name);
        log(" <-- data: " + dump(result));
        return result;
    }

    public ObjectData resolveImmediately(final Session session, final IdentityData target) {
        log("resolve immediately" + dump(target));
        ObjectData result = decorated.resolveImmediately(session, target);
        log("  <-- data: " + dump(result));
        return result;

    }

    public ObjectData[] setAssociation(
            final Session session,
            final String fieldIdentifier,
            final IdentityData target,
            final IdentityData associate) {
        log("set association " + fieldIdentifier + indentedNewLine() + "target: " + dump(target) + indentedNewLine()
                + "associate: " + dump(associate));
        ObjectData[] changes = decorated.setAssociation(session, fieldIdentifier, target, associate);
        log("  <-- changes: " + dump(changes));
        return changes;
    }

    public ObjectData[] setValue(
            final Session session,
            final String fieldIdentifier,
            final IdentityData target,
            final ValueData value) {
        log("set value " + fieldIdentifier + indentedNewLine() + "target: " + dump(target) + indentedNewLine() + "value: "
                + value);
        ObjectData[] changes = decorated.setValue(session, fieldIdentifier, target, value);
        log("  <-- changes: " + dump(changes));
        return changes;
    }

    public void shutdown() {
        decorated.shutdown();
    }
}
// Copyright (c) Naked Objects Group Ltd.
