package org.nakedobjects.nof.reflect.spec;

import org.apache.log4j.Logger;
import org.nakedobjects.noa.reflect.NakedObjectReflector;
import org.nakedobjects.noa.spec.Features;
import org.nakedobjects.noa.spec.NakedObjectSpecification;
import org.nakedobjects.nof.core.context.Perspective;
import org.nakedobjects.nof.core.util.Assert;
import org.nakedobjects.nof.core.util.DebugInfo;
import org.nakedobjects.nof.core.util.DebugString;
import org.nakedobjects.nof.reflect.peer.ReflectionPeerFactory;
import org.nakedobjects.nof.reflect.perspective.PerspectiveSpecification;


public abstract class AbstractReflector implements NakedObjectReflector, DebugInfo {
    private static final String ARRAY_PREFIX = "[";
    private static final String DOT_NET_PACKAGES_PREFIX = "System.";
    private static final String JAVA_PACKAGES_PREFIX = "java.";
    private final static Logger LOG = Logger.getLogger(AbstractReflector.class);
    private SpecificationCache cache;
    private ReflectionPeerBuilder reflectionPeerBuilder;

    /**
     * Return all the loaded specifications.
     */
    public NakedObjectSpecification[] allSpecifications() {
        return getCache().allSpecifications();
    }

    protected abstract boolean collectionClass(Class cls, String className);

    protected Class collectionElementType() {
        return Object.class;
    }

    public void debugData(final DebugString str) {
        reflectionPeerBuilder.debugData(str);

        str.appendTitle("Specifications");
        NakedObjectSpecification[] specs = allSpecifications();
        for (int i = 0; i < specs.length; i++) {
            NakedObjectSpecification specification = specs[i];
            str.append(Features.isAbstract(specification) ? "A" : ".");
            str.append(Features.isAggregated(specification) ? "G" : ".");
            str.append(Features.isService(specification) ? "S" : ".");
            str.append(Features.isBoundedSet(specification) ? "B" : ".");
            int type = specification.getType();
            str.append(type == NakedObjectSpecification.COLLECTION ? "C" : ".");
            str.append(type == NakedObjectSpecification.OBJECT ? "O" : ".");
            str.append(type == NakedObjectSpecification.VALUE ? "V" : ".");
            str.append("  ");
            str.append(specification.getShortName());
            str.append("  [fqc=");
            str.append(specification.getFullName());
            str.append(",type=");
            str.append(specification.getClass().getName());
            str.appendln("]");
        }
    }

    public String debugTitle() {
        return "Reflector";
    }

    protected boolean ignoreClass(Class cls, String className) {
        return false;
    }

    protected boolean ignoreDotNetClass(String className) {
        return true;
    }

    protected boolean ignoreJavaClass(Class cls, String className) {
        return false;
    }

    public void init() {
        LOG.debug("initialising " + this);
        Assert.assertNotNull("ReflectionPeerBuilder needed", reflectionPeerBuilder);
        reflectionPeerBuilder.init();
    }

    protected abstract AbstractSpecification install(final Class cls, final ReflectionPeerBuilder builder);

    public void installServiceSpecification(Class cls) {
        NakedObjectSpecification spec = getCache().get(cls.getName());
        if (spec != null) {
            return;
        }

        AbstractSpecification specification = install(cls, reflectionPeerBuilder);
        getCache().cache(cls.getName(), specification);
        specification.introspect(reflectionPeerBuilder);
        specification.markAsService();
    }

    private NakedObjectSpecification load(final Class cls) {
        synchronized (getCache()) {
            String className = cls.getName();
            AbstractSpecification specification;
            if (Perspective.class.isAssignableFrom(cls)) {
                specification = new PerspectiveSpecification();

            } else if (primitiveClass(cls, className)) {
                PrimitiveSpecification spec = new PrimitiveSpecification(cls);
                getCache().cache(cls.getName(), spec);
                return spec;

            } else if (ignoreClass(cls, className)) {
                specification = new NoMemberSpecification(className);

            } else if (className.startsWith(ARRAY_PREFIX)) {
                specification = new CollectionSpecification(className, cls.getComponentType());

            } else if (collectionClass(cls, className)) {
                specification = new CollectionSpecification(className, collectionElementType());

            } else if (className.startsWith(JAVA_PACKAGES_PREFIX) && ignoreJavaClass(cls, className)) {
                specification = new NoMemberSpecification(className);

            } else if (className.startsWith(DOT_NET_PACKAGES_PREFIX) && ignoreDotNetClass(className)) {
                specification = new NoMemberSpecification(className);

            } else {
                specification = install(cls, reflectionPeerBuilder);
            }

            if (specification == null) {
                LOG.info("unrecognised class " + className + "; 'null' specification created");
                specification = new NoMemberSpecification(className);
            }

            getCache().cache(cls.getName(), specification);
            specification.introspect(reflectionPeerBuilder);
            return specification;
        }
    }

    /**
     * Return the specification for the specified class of object.
     */
    public NakedObjectSpecification loadSpecification(Class cls) {
        Assert.assertNotNull(cls);

        NakedObjectSpecification spec = getCache().get(cls.getName());
        if (spec != null) {
            return spec;
        }

        if (cls.getName().endsWith("Value")) {
            synchronized (getCache()) {
                spec = new ValueSpecification(cls);
                getCache().cache(cls.getName(), spec);
            }
            return spec;
        }
        if (cls.isPrimitive()) {
            synchronized (getCache()) {
                spec = new PrimitiveSpecification(cls);
                getCache().cache(cls.getName(), spec);
            }
            return spec;
        } else {
            return load(cls);
        }

    }

    /**
     * Return the specification for the specified class of object.
     */
    public final NakedObjectSpecification loadSpecification(String className) {
        Assert.assertNotNull("specification class must be specified", className);

        try {
            Class cls = Class.forName(className);
            return loadSpecification(cls);
        } catch (ClassNotFoundException e) {
            NakedObjectSpecification spec = getCache().get(className);
            if (spec != null) {
                return spec;
            }
            
            LOG.warn("not a class " + className + "; 'null' specification created");
            spec = new NoMemberSpecification(className);
            getCache().cache(className, spec);
            return spec;
        }

    }

    protected abstract boolean primitiveClass(Class cls, String className);

    /**
     * Expose as a .NET property
     * 
     * @property
     */
    public void set_Cache(final SpecificationCache cache) {
        setCache(cache);
    }

    /**
     * Expose as a .NET property
     * 
     * @property
     */
    public void set_ReflectionPeerFactories(final ReflectionPeerFactory[] factories) {
        setReflectionPeerFactories(factories);
    }

    public void setCache(final SpecificationCache cache) {
        this.cache = cache;
    }

    public void setReflectionPeerFactories(final ReflectionPeerFactory[] factories) {
        reflectionPeerBuilder = new ReflectionPeerBuilder();
        reflectionPeerBuilder.setFactories(factories);
    }

    public void shutdown() {
        LOG.info("shutting down " + this);
        getCache().clear();
        reflectionPeerBuilder.shutdown();
    }

    private SpecificationCache getCache() {
        if (cache == null) {
            cache = new SimpleSpecificationCache();
        }
        return cache;
    }

}
// Copyright (c) Naked Objects Group Ltd.
