package org.nakedobjects.nof.reflect.spec;

import org.nakedobjects.noa.NakedObjectRuntimeException;
import org.nakedobjects.noa.adapter.Naked;
import org.nakedobjects.noa.adapter.NakedObject;
import org.nakedobjects.noa.adapter.NakedReference;
import org.nakedobjects.noa.reflect.Consent;
import org.nakedobjects.noa.reflect.OneToManyAssociation;
import org.nakedobjects.noa.spec.NakedObjectSpecification;
import org.nakedobjects.nof.core.context.NakedObjectsContext;
import org.nakedobjects.nof.core.reflect.Veto;
import org.nakedobjects.nof.core.util.DebugString;
import org.nakedobjects.nof.core.util.ToString;
import org.nakedobjects.nof.reflect.peer.OneToManyPeer;


public class OneToManyAssociationImpl extends AbstractNakedObjectField implements OneToManyAssociation {
    private final OneToManyPeer reflectiveAdapter;

    public OneToManyAssociationImpl(final String methodId, final NakedObjectSpecification type, final OneToManyPeer association) {
        super(methodId, type);
        this.reflectiveAdapter = association;
    }

    public void addElement(final NakedObject inObject, final NakedObject associate) {
        if (associate == null) {
            throw new IllegalArgumentException("Can't use null to add an item to a collection");
        }
        if (readWrite()) {
            if (inObject.getResolveState().isPersistent() && associate.getResolveState().isTransient()) {
                throw new NakedObjectRuntimeException("can't set a reference to a transient object from a persistent one: "
                        + inObject.titleString() + " (persistent) -> " + associate.titleString() + " (transient)");
            }
            reflectiveAdapter.addAssociation(inObject, associate);
        }
    }

    public void clearCollection(final NakedObject inObject) {
        if (readWrite()) {
            reflectiveAdapter.removeAllAssociations(inObject);
        }
    }

    public String debugData() {
        DebugString debugString = new DebugString();
        debugString.indent();
        debugString.indent();
        reflectiveAdapter.debugData(debugString);
        return debugString.toString();
    }

    public Naked get(final NakedObject fromObject) {
        return reflectiveAdapter.getAssociations(fromObject);
    }

    public String getBusinessKeyName() {
        return reflectiveAdapter.getBusinessKeyName();
    }

    public String getDescription() {
        return reflectiveAdapter.getDescription();
    }

    public Object getExtension(final Class cls) {
        return reflectiveAdapter.getExtension(cls);
    }

    public Class[] getExtensions() {
        return reflectiveAdapter.getExtensions();
    }

    public String getHelp() {
        return reflectiveAdapter.getHelp();
    }

    /**
     * Return the default label for this member. This is based on the name of this member.
     * 
     * @see #getId()
     */
    public String getName() {
        String label = reflectiveAdapter.getName();
        return label == null ? defaultName : label;
    }

    public void initCollection(final NakedObject inObject, final NakedObject[] instances) {
        if (readWrite()) {
            reflectiveAdapter.initOneToManyAssociation(inObject, instances);
        }
    }

    public boolean isCollection() {
        return true;
    }

    public boolean isPersisted() {
        return reflectiveAdapter.isPersisted();
    }

    public boolean isEmpty(final NakedObject inObject) {
        return reflectiveAdapter.isEmpty(inObject);
    }

    public boolean isOptionEnable() {
        return false;
    }

    public Consent isUsableDeclaratively() {
        return reflectiveAdapter.isUsableDeclaratively();
    }

    public Consent isUsableForSession() {
        return reflectiveAdapter.isUsableForSession(NakedObjectsContext.getSession());
    }


    public Consent isUsable(final NakedReference target) {
        return reflectiveAdapter.isUsable(target);
    }

    public Consent isValidToAdd(final NakedObject container, final NakedObject element) {
        if(!readWrite()) {
            return new Veto("Can't add to a collection that is not persisted");
        }
        return reflectiveAdapter.isAddValid(container, element);
    }

    public Consent isValidToRemove(final NakedObject container, final NakedObject element) {
        if(!readWrite()) {
            return new Veto("Can't remove from a collection that is not persisted");
        }
        return reflectiveAdapter.isRemoveValid(container, element);
    }

    public boolean isVisibleDeclaratively() {
        return reflectiveAdapter.isVisibleDeclaratively();
    }
    
    public boolean isVisibleForSession() {
        return reflectiveAdapter.isVisibleForSession(NakedObjectsContext.getSession());
    }
    
    public boolean isVisible(final NakedReference target) {
        return reflectiveAdapter.isVisible(target);
    }

    private boolean readWrite() {
        return reflectiveAdapter.isPersisted();
    }

    public void removeElement(final NakedObject inObject, final NakedObject associate) {
        if (associate == null) {
            throw new IllegalArgumentException("element should not be null");
        }
        if (readWrite()) {
            reflectiveAdapter.removeAssociation(inObject, associate);
        }
    }

    public Object getDefault(
            NakedObject nakedObject) {
        return null;
    }

    public void toDefault(NakedObject target) {
    }

    public String toString() {
        ToString str = new ToString(this);
        str.append(super.toString());
        str.append(",");
        str.append("persisted", isPersisted());
        str.append("type", getSpecification() == null ? "unknown" : getSpecification().getShortName());
        return str.toString();
    }


}
// Copyright (c) Naked Objects Group Ltd.
