package org.nakedobjects.nof.reflect.spec;

import java.util.Enumeration;

import org.nakedobjects.noa.NakedObjectRuntimeException;
import org.nakedobjects.noa.adapter.Naked;
import org.nakedobjects.noa.adapter.NakedCollection;
import org.nakedobjects.noa.adapter.NakedObject;
import org.nakedobjects.noa.adapter.NakedReference;
import org.nakedobjects.noa.adapter.NakedValue;
import org.nakedobjects.noa.reflect.Consent;
import org.nakedobjects.noa.reflect.OneToOneAssociation;
import org.nakedobjects.noa.spec.Features;
import org.nakedobjects.noa.spec.NakedObjectSpecification;
import org.nakedobjects.nof.core.context.NakedObjectsContext;
import org.nakedobjects.nof.core.persist.AllInstances;
import org.nakedobjects.nof.core.util.DebugString;
import org.nakedobjects.nof.core.util.ToString;
import org.nakedobjects.nof.reflect.peer.OneToOnePeer;


public class OneToOneAssociationImpl extends AbstractNakedObjectField implements OneToOneAssociation {

    private final OneToOnePeer reflectiveAdapter;

    public OneToOneAssociationImpl(
            final String fieldId,
            final NakedObjectSpecification specification,
            final OneToOnePeer association) {
        super(fieldId, specification);
        this.reflectiveAdapter = association;
    }

    public void clearAssociation(final NakedObject inObject, final NakedObject associate) {
        if (associate == null) {
            throw new NullPointerException("Must specify the item to remove/dissociate");
        }
        reflectiveAdapter.clearAssociation(inObject, associate);
    }

    public void clearValue(final NakedObject inObject) {
        NakedValue associate = (NakedValue) get(inObject);
        if (associate != null) {
            associate.clear();
        }
    }

    public String debugData() {
        DebugString debugString = new DebugString();
        debugString.indent();
        debugString.indent();
        reflectiveAdapter.debugData(debugString);
        return debugString.toString();
    }

    public Naked get(final NakedObject fromObject) {
        return reflectiveAdapter.getAssociation(fromObject);
    }

    public String getBusinessKeyName() {
        return reflectiveAdapter.getBusinessKeyName();
    }

    public String getDescription() {
        return reflectiveAdapter.getDescription();
    }

    public Object getExtension(final Class cls) {
        return reflectiveAdapter.getExtension(cls);
    }

    public Class[] getExtensions() {
        return reflectiveAdapter.getExtensions();
    }

    public String getHelp() {
        return reflectiveAdapter.getHelp();
    }

    /**
     * Return the default label for this member. This is based on the name of this member.
     * 
     * @see #getId()
     */
    public String getName() {
        String label = reflectiveAdapter.getName();
        return label == null ? defaultName : label;
    }

    public Naked[] getOptions(NakedObject target) {
        Object[] pojoOptions = reflectiveAdapter.getOptions(target);
        if (pojoOptions != null) {
            NakedObject[] options = new NakedObject[pojoOptions.length];
            for (int i = 0; i < options.length; i++) {
                options[i] = NakedObjectsContext.getObjectLoader().getAdapterFor(pojoOptions[i]);
            }
            return options;
        } else if (Features.isBoundedSet(getSpecification())) {
            NakedCollection instances = NakedObjectsContext.getObjectPersistor().findInstances(
                    new AllInstances(getSpecification(), false));
            NakedObject[] options = new NakedObject[instances.size()];
            int j = 0;
            for (Enumeration e = instances.elements(); e.hasMoreElements();) {
                options[j++] = ((NakedObject) e.nextElement());
            }
            return options;
        }
        return null;
    }

    public void initAssociation(final NakedObject inObject, final NakedObject associate) {
        if (isReadWrite()) {
            reflectiveAdapter.initAssociation(inObject, associate);
        }
    }

    public Consent isAssociationValid(final NakedObject inObject, final NakedObject value) {
        return reflectiveAdapter.isAssociationValid(inObject, value);
    }

    public boolean isPersisted() {
        return reflectiveAdapter.isPersisted();
    }

    public boolean isEmpty(final NakedObject inObject) {
        return reflectiveAdapter.isEmpty(inObject);
    }

    public boolean isMandatory() {
        return reflectiveAdapter.isMandatory();
    }

    public boolean isObject() {
        return true;
    }

    public boolean isOptionEnable() {
        return Features.isBoundedSet(getSpecification()) || reflectiveAdapter.isOptionEnabled();
    }

    public Consent isUsableDeclaratively() {
        return reflectiveAdapter.isUsableDeclaratively();
    }

    public Consent isUsableForSession() {
        return reflectiveAdapter.isUsableForSession(NakedObjectsContext.getSession());
    }

    public Consent isUsable(final NakedReference inObject) {
        return reflectiveAdapter.isUsable(inObject);
    }

    public boolean isVisibleDeclaratively() {
        return reflectiveAdapter.isVisibleDeclaratively();
    }

    public boolean isVisibleForSession() {
        return reflectiveAdapter.isVisibleForSession(NakedObjectsContext.getSession());
    }

    public boolean isVisible(final NakedReference target) {
        return reflectiveAdapter.isVisible(target);
    }

    private boolean isReadWrite() {
        // TODO confirm there is no cases when a reference should not be set up up, and remove this method.
        return true;
    }

    public Object getDefault(
            NakedObject nakedObject) {
        return (NakedObject) reflectiveAdapter.getDefault(nakedObject);
    }

    public void toDefault(final NakedObject inObject) {
        NakedObject defaultValue = (NakedObject) getDefault(inObject);
        if (defaultValue != null) {
            setAssociation(inObject, defaultValue);
        }
    }

    public void setAssociation(final NakedObject inObject, final NakedObject associate) {
        if (isReadWrite()) {
            if (inObject.getResolveState().isPersistent() && associate != null && associate.getResolveState().isTransient()) {
                throw new NakedObjectRuntimeException("can't set a reference to a transient object from a persistent one: "
                        + inObject.titleString() + " (persistent) -> " + associate.titleString() + " (transient)");
            }
            reflectiveAdapter.setAssociation(inObject, associate);
        }
    }

    public String toString() {
        ToString str = new ToString(this);
        str.append(super.toString());
        str.setAddComma();
        str.append("persisted", isPersisted());
        str.append("type", getSpecification().getShortName());
        return str.toString();
    }

}
// Copyright (c) Naked Objects Group Ltd.
