package org.nakedobjects.nof.reflect.spec;

import java.util.HashMap;
import java.util.Map;

import org.nakedobjects.noa.adapter.NakedObject;
import org.nakedobjects.noa.adapter.NakedReference;
import org.nakedobjects.noa.adapter.Persistable;
import org.nakedobjects.noa.reflect.Consent;
import org.nakedobjects.noa.reflect.NakedObjectAction;
import org.nakedobjects.noa.reflect.NakedObjectField;
import org.nakedobjects.noa.reflect.NakedObjectField.Filter;
import org.nakedobjects.noa.spec.NakedObjectSpecification;
import org.nakedobjects.nof.core.reflect.Allow;
import org.nakedobjects.nof.core.util.DebugString;
import org.nakedobjects.nof.core.util.ToString;
import org.nakedobjects.nof.core.util.UnexpectedCallException;


/**
 * A simple implementation used for objects that have no members (fields or actions). Used for primitives and
 * as a fall-back if no specification can be otherwise generated.
 */
public class PrimitiveSpecification implements NakedObjectSpecification {
    private final Class underlyingClass;
    private final String name;

    
    public PrimitiveSpecification(Class cls) {
        this.underlyingClass = cls;
        this.name = underlyingClass.getName();
        addExtension(underlyingClass);
    }

    
    public void addSubclass(final NakedObjectSpecification specification) {}

    public void clearDirty(final NakedObject object) {}

    public void debugData(DebugString debug) {
        debug.append("There are no reflective actions");
    }

    public String debugTitle() {
        return "Java Primitive Specification";
    }
   
    public NakedObjectField[] getStaticallyVisibleFields() {
        return new NakedObjectField[0];
    }

    public NakedObject getAggregate(NakedObject object) {
        throw new UnexpectedCallException("Can't deal with aggregated objects");
    }
    
    public NakedObjectAction getClassAction(
            final NakedObjectAction.Type type, final String name, final NakedObjectSpecification[] parameters) {
        return null;
    }

    public NakedObjectAction[] getClassActions(final NakedObjectAction.Type type) {
        return new NakedObjectAction[0];
    }

    
    ////////////////////// Extensions ////////////////////
    
    private Map extensionsByClass = new HashMap();

    /**
     * Can obtain the underlying class used to build this specification.
     * 
     * <p>
     * To do this, use (rather confusingly) <tt>Class.class</tt> as the key:
     * <pre>
     * Class underlyingClass = someJavaSpec.getExtension(Class.class);
     * </pre>
     */
    public Object getExtension(Class cls) {
        return extensionsByClass.get(cls);
    }
    
    public Class[] getExtensions() {
        return (Class[]) extensionsByClass.keySet().toArray();
    }
    
    /**
     * Not part of the API.
     * 
     * <p>
     * However subclasses such as <tt>JavaSpecification</tt> promote this to
     * an (internal) API through implementing <tt>ExtensionHolderMutable</tt>.
     * 
     * @param extension
     */
    protected void addExtension(Object extension) {
        extensionsByClass.put(extension.getClass(), extension);
    }


    
    
    
    public int getFeatures() {
        return 0;
    }

    public NakedObjectField getField(final String name) {
        return null;
    }

    public Object getFieldExtension(final String name, final Class cls) {
        return null;
    }

    public Class[] getFieldExtensions(final String name) {
        return new Class[0];
    }

    public NakedObjectField[] getFields() {
        return new NakedObjectField[0];
    }

    public NakedObjectField[] getFields(
            Filter filter) {
        return new NakedObjectField[0];
    }

    public String getFullName() {
        return name;
    }

    public String getIconName(NakedReference reference) {
        return null;
    }
    
    public NakedObjectAction getObjectAction(
            final NakedObjectAction.Type type, final String name, final NakedObjectSpecification[] parameters) {
        return null;
    }

    public NakedObjectAction[] getObjectActions(final NakedObjectAction.Type type) {
        return new NakedObjectAction[0];
    }

    public String getPluralName() {
        return name;
    }

    public Class getSearchViaRepository() {
        return null;
    }
    
    public String getShortName() {
        return name.substring(name.lastIndexOf('.') + 1);
    }

    public String getSingularName() {
        return name.substring(name.lastIndexOf('.') + 1);
    }

    public String getDescription() {
        return getSingularName();
    }


    public String getTitle(final NakedObject naked) {
        return "no title";
    }

    public int getType() {
        return VALUE;
    }
    
    public NakedObjectField[] getDynamicallyVisibleFields(final NakedObject object) {
        return new NakedObjectField[0];
    }

    public boolean hasSubclasses() {
        return false;
    }

    public NakedObjectSpecification[] interfaces() {
        return new NakedObjectSpecification[0];
    }

    public void introspect() {}

    public boolean isDirty(final NakedObject object) {
        return false;
    }

    public boolean isOfType(final NakedObjectSpecification specification) {
        return specification == this;
    }

    public Consent isPersistable(NakedObject transientObject) {
        return Allow.DEFAULT;
    }

    public void lifecycleEvent(NakedObject object, int event) {}

    public void markDirty(final NakedObject object) {}
    
    public Persistable persistable() {
        return Persistable.TRANSIENT;
    }

    public boolean queryByExample() {
        return false;
    }

    public NakedObjectSpecification[] subclasses() {
        return new NakedObjectSpecification[0];
    }

    public NakedObjectSpecification superclass() {
        return null;
    }

    public String toString() {
        ToString str = new ToString(this);
        str.append("class", name);
        return str.toString();
    }

    private static Map defaultValueByClass = new HashMap();
    static {
        defaultValueByClass.put(Byte.class, new Byte((byte)0));
        defaultValueByClass.put(byte.class, new Byte((byte)0));
        defaultValueByClass.put(Short.class, new Short((short)0));
        defaultValueByClass.put(short.class, new Short((short)0));
        defaultValueByClass.put(Integer.class, new Integer(0));
        defaultValueByClass.put(int.class, new Integer(0));
        defaultValueByClass.put(Long.class, new Long(0L));
        defaultValueByClass.put(long.class, new Long(0L));
        defaultValueByClass.put(Float.class, new Float(0.0f));
        defaultValueByClass.put(float.class, new Float(0.0f));
        defaultValueByClass.put(Double.class, new Double(0.0));
        defaultValueByClass.put(double.class, new Double(0.0));
        defaultValueByClass.put(Character.class, new Character(' '));
        defaultValueByClass.put(char.class, new Character(' '));
        defaultValueByClass.put(Boolean.class, Boolean.FALSE);
        defaultValueByClass.put(Boolean.class, Boolean.FALSE);
        defaultValueByClass.put(String.class, "");
        defaultValueByClass.put(java.util.Date.class, new java.util.Date(0)); // TODO: what should be the default time?
    }

    private boolean checkedForDefault = false;
    private Object defaultValue;
    public Object getDefaultValue() {
        if (!checkedForDefault) {
            defaultValue = defaultValueByClass.get(underlyingClass);
            checkedForDefault = true;
        }
        return defaultValue;
    }


    
}
// Copyright (c) Naked Objects Group Ltd.
